
/*****************************************************************************
 *                                                                           *
 *  PMB source code, MPI-1 part                           ---/---            *
 *                                                          / /              *
 *  Copyright (c) PALLAS GmbH 1998                         / / /             *
 *                                                        / / / /            *
 *                                                         / / /             *
 *  Authors: Hans-Joachim Plum , Gero Ritzenhoefer          / / PALLAS       *
 *  Revision: 2.0                                         ---/---            *
 *  Date: 1998/03/10                                                         *
 *                                                                           *
 *****************************************************************************/
/*-------------------------------------------------------------------
Initialization functions

Functions implemented in this module: 

void Basic_Input      (struct comm_info* c_info, struct Bench** P_BList,
                      int *argc, char ***argv, int* NP_min)
void Init_pointers    (struct comm_info* c_info)
int  Init_communicator(struct comm_info* c_info, int NP   )
void Set_Communicator (struct comm_info* c_info           )
void Init_buffers     (struct comm_info* c_info, struct Bench* Bmark, int size)
int  valid            (struct comm_info* c_info, struct Bench* Bmark, int NP)
void Free_ALL         (struct comm_info* c_info, struct Bench** P_BList)

---------------------------------------------------------------------*/
#include "declare.h"
#include "Benchmark.h"

/* EXTERN PROTOTYPES */
extern void General_Info();
extern void Construct_BList
         (struct Bench** P_BList, char* name);
extern void Show_procids( struct comm_info *c_info );
extern void Err_Hand(int,int);

/* LOCAL PROTOTYPES */
void Basic_Input      (struct comm_info* c_info, struct Bench** P_BList,
                      int *argc, char ***argv, int* NP_min);
void Init_pointers    (struct comm_info* c_info);
int  Init_communicator(struct comm_info* c_info, int NP   );
void Set_Communicator (struct comm_info* c_info           );
int  valid            (struct comm_info* c_info, struct Bench* Bmark, int NP);
void Init_buffers     (struct comm_info* c_info, struct Bench* Bmark, int size);

/********************************************************************/
void Basic_Input(struct comm_info* c_info,
                 struct Bench** P_BList,
                 int *argc, char ***argv,int* NP_min)
/********************************************************************


---------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
---------------------------------------------------------------------
Input:       argc     | int*              | number of comm. line args
             argv     | char***           | command line arguments
             
In/Out     : c_info   | struct comm_info* | see comm_info.h 
                      |                   | (will be partially initialized)
                      |                   |
Output     : 
             P_BList  | struct Bench**    | *P_Blist = 
                      |                   | array of Benchmarks descriptors
                      |                   |
             NP_min   | int               | minimum number of processes
                      |                   |  
----------------------------------------------------------------------
----------------------------------------------------------------------
Description: Initializing options defined in command line.
             1. Selection of individual benchmarks       
	        (e.g. mpirun -np * pmb Allreduce ) 
             2. Selection of additional args:
	        (e.g. mpirun -np * pmb -npmin * -group * ) 
--------------------------------------------------------------------*/


{
  int i,n_cases,add_args, iarg;

  char* new_file_name;

  int length = strlen(OUTPUT_FILENAME);

  MPI_Comm_rank(MPI_COMM_WORLD,&c_info->w_rank);
  MPI_Comm_size(MPI_COMM_WORLD,&c_info->w_num_procs);
  
  unit = stdout; 
  if( c_info->w_rank == 0 && length > 0 )
    {

      new_file_name = (char*) malloc( length + 5);
      sprintf(new_file_name,"%s_%d",OUTPUT_FILENAME,c_info->w_num_procs);
      unit = fopen(new_file_name,"w"); 


    }

  c_info->group_mode = -1;
  *NP_min=2;
	    
  Construct_BList(P_BList,"ALL");
 
}

/********************************************************************/
void Init_pointers(struct comm_info *c_info )
{
/********************************************************************


---------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
---------------------------------------------------------------------
In/Out     : c_info   | struct comm_info* | see comm_info.h 
                      |                   | Pointers initialized
----------------------------------------------------------------------*/

MPI_Comm_size(MPI_COMM_WORLD,&c_info->w_num_procs);
MPI_Comm_rank(MPI_COMM_WORLD,&c_info->w_rank     );

c_info->s_data_type   = MPI_BYTE;  /* DATA TYPE of SEND    BUFFER    */ 
c_info->r_data_type   = MPI_BYTE;  /* DATA TYPE of RECEIVE BUFFER    */

c_info->op_type       = MPI_SUM;   /* OPERATION TYPE IN Allred       */
c_info->red_data_type = MPI_FLOAT; /* NOTE: NO 'CAST' CHECK IN. IBUF */

c_info -> s_buffer = c_info -> r_buffer = NULL;
c_info -> s_alloc  = c_info -> r_alloc = 0;

c_info->communicator= MPI_COMM_NULL;

/* Auxiliary space */
c_info->g_ranks = (int*) malloc( c_info->w_num_procs * sizeof(int) );
c_info->g_sizes = (int*) malloc( c_info->w_num_procs * sizeof(int) );


c_info->reccnt  = (int*) malloc( c_info->w_num_procs * sizeof(int) );
c_info->displs  = (int*) malloc( c_info->w_num_procs * sizeof(int) );

if( c_info->g_ranks == NULL || c_info->g_sizes == NULL ||
    c_info->reccnt  == NULL || c_info->displs  == NULL ) 
  Err_Hand(0, APPL_ERR_MALLOC);


}


/**********************************************************************/
int Init_communicator(struct comm_info* c_info, int NP)
/**********************************************************************/

/*-------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
---------------------------------------------------------------------
                      |                   |
Input      : NP       | int               | number of processes to
                      |                   | run the benchmark (may
                      |                   | be less than size of
                      |                   | MPI_COMM_WORLD)
                      |                   |
In/Out     : c_info   | struct comm_info* | see comm_info.h 
                      |                   | communicator initialized
                      |                   | group management set up 
----------------------------------------------------------------------*/

{
int ir,k,i,ierr,snd,cnt,proc,*aux_ptr,aux;
int color,key;

MPI_Group group, w_group;
MPI_Status stat;

c_info->NP=NP;                     /* NUMBER OF OVERALL PROCESSES    */          
Set_Communicator( c_info );        /* GROUP MANAGEMENT               */

/* INITIALIZATION  WITHIN THE ACTUAL COMMUNICATOR */
if( c_info->communicator != MPI_COMM_NULL )
  {
    MPI_Comm_size(c_info->communicator,&(c_info ->num_procs));
    MPI_Comm_rank(c_info->communicator,&(c_info ->rank));
    
    c_info->pair0 = 0;
    c_info->pair1 = c_info ->num_procs-1;
    
    c_info->select_tag = 0;
    c_info->select_source = 0;
  }
else
  {
    c_info -> rank = -1;
  }

if( c_info->communicator == MPI_COMM_WORLD ) 
  {
    c_info->n_groups = 1;
    c_info->g_sizes[0] = c_info->w_num_procs;
    for(i=0; i<c_info->w_num_procs; i++) c_info->g_ranks[i]=i;
  }
else
  {
    /* Collect global group information */
    if( c_info->rank == 0 )
      {
	/* group leaders provide group ranks */
	MPI_Comm_group(MPI_COMM_WORLD,&w_group);
	MPI_Comm_group(c_info->communicator,&group);
	for (i=0; i<c_info->num_procs; i++) c_info->g_sizes[i] = i;

	/* TRANSLATION OF RANKS */
	MPI_Group_translate_ranks( group, c_info->num_procs, 
				   c_info->g_sizes,w_group,
				   c_info->g_ranks );
	snd = c_info->num_procs;
      }
    else
      {
	*c_info->g_ranks = -1;
	snd = 1;
      }
    /* w_rank 0 collects in g_ranks ranks of single groups */
    if( c_info->w_rank == 0 ) 
      {
        if( c_info->rank == 0 )
	  {
            c_info->n_groups = 1;
            c_info->g_sizes[c_info->n_groups-1] = c_info->num_procs;
            aux_ptr = c_info->g_ranks+c_info->num_procs;
	  }
	else
	  {
            c_info->n_groups = 0;
            aux_ptr=c_info->g_ranks;
	  }
	for( proc=1; proc<c_info->w_num_procs; proc++ )
	  {
	    /* Recv group ranks or -1  */
	    cnt = (int)(c_info->g_ranks+c_info->w_num_procs-aux_ptr);
	    if( cnt <= 0 )
	      /* all leaders have sent! */
	      {
		aux_ptr = &i;
		cnt=1;
	      }
	    MPI_Recv(aux_ptr,cnt,MPI_INT,proc,1000,MPI_COMM_WORLD,&stat);
	    
	    if( *aux_ptr >= 0 ) 
	      {
		/* Message was from a group leader  */
		c_info->n_groups ++;
		MPI_Get_count( &stat, MPI_INT,
			       &c_info->g_sizes[c_info->n_groups-1]);
		aux_ptr += c_info->g_sizes[c_info->n_groups-1];
	      }
	  }
      }
    else  /* w_rank != 0 */
      {
	MPI_Send(c_info->g_ranks,snd,MPI_INT,0,1000,MPI_COMM_WORLD);
      }
  }
/* End collection of group information */   
return 0;
}


/**********************************************************************/
void Set_Communicator( struct comm_info *c_info )
/**********************************************************************

----------------------------------------------------------------------
             VARIABLE              |       TYPE        |   MEANING
----------------------------------------------------------------------
Input      :                       |                   |  
Output     :                       |                   |
                                   |                   |  
In/Out     :  c_info               | struct comm_info* | see comm_info.h
                                   |                   | Communicator set up
-----------------------------------------------------------------------

-----------------------------------------------------------------------
Description: Initialization of communicators by group coloring
---------------------------------------------------------------------*/
{
  int color,key,i;
  
  /* insert choice for communicator here;
     NOTE   :  globally more than one communicator is allowed   
     Example: grouping of pairs of processes: 
     0 0 1 1 2 2  .. (if even),  UNDEF 0 0 1 1 2 2  .. (if odd) 
     */

  if( c_info->communicator != MPI_COMM_NULL && 
      c_info->communicator != MPI_COMM_WORLD)
  MPI_Comm_free(&c_info->communicator);


  /* nearest  neighbour groups (0,1),(2,3), ... */

  if(c_info->group_mode == 1)
    {
      i=c_info->w_rank;
      color = i/c_info->NP;
      key = 0;
      if(color >= c_info->w_num_procs/c_info->NP) color=MPI_UNDEFINED;
      MPI_Comm_split(MPI_COMM_WORLD, color, key, &c_info->communicator);
    }

  /* groups (0, w_num_procs-1), (1, w_num_procs-2), ..., (w_num_procs/2-1, )*/ 

    else if (c_info->group_mode == 2)  
      {
	i=c_info->w_rank;
      
	if(i < c_info->w_num_procs/2)
	  color = i;
	else
	  color = c_info->w_num_procs -1 -i;

	key = 0;
	if(color >= c_info->w_num_procs/c_info->NP) color=MPI_UNDEFINED;

	MPI_Comm_split(MPI_COMM_WORLD, color, key, &c_info->communicator);

      }

  /* Default choice and Group definition.  */


  else
    {
      if(c_info->w_rank < c_info->NP) color=0;
      else color=MPI_UNDEFINED;   
      key=0;
      MPI_Comm_split(MPI_COMM_WORLD, color, key, &c_info->communicator); 
    }
}

int valid(struct comm_info * c_info, struct Bench* Bmark, int NP)
{
/* Checks validity of Bmark for NP processes */
/* Erroneous cases: */
int invalid, skip;
invalid = Bmark->type == SingleTransfer && NP <= 1;
skip    = Bmark->type == SingleTransfer && NP > 2;

if ( invalid )
  {
  if( c_info->w_rank == 0 )
  {
  if( NP == 1 )
  fprintf(unit,"\n# !! Benchmark %s invalid for %d process   !! \n\n",Bmark->name,NP);
  else
  fprintf(unit,"\n# !! Benchmark %s invalid for %d processes !! \n\n",Bmark->name,NP);
  }
 
  return 0;
  }

/* Cases to skip: */
if ( skip ) return 0;

return 1;
}


/********************************************************************/
void Init_buffers(struct comm_info* c_info, struct Bench* Bmark, int size)
/********************************************************************


             VARIABLE |       TYPE        |   MEANING
--------------------------------------------------------------------
Input      : Bmark    | struct Bench*     | Benchmark to run
             size     | int               | size of part to be initialized
             new      | int               | logical: new allocation?
Output     :          |                   |  
In/Out     : c_info   |struct comm_info*  | see comm_info.c
                      |                   | Comm. buffers initialized
---------------------------------------------------------------------*/
{
  int s_len, r_len, len, maxlen;
  int i;

  Type_Size s_size;

  if(c_info->rank < 0 ) return;

  maxlen = 1<<MAXMSGLOG;         /* allocate buffers maxlen=2^MAXMSGLOG */

  if( !strcmp(Bmark->name,"Allgather") || !strcmp(Bmark->name,"Allgatherv") )
    {
      s_len = maxlen;
      r_len = c_info->num_procs*maxlen;
    }
  else
      s_len = r_len = maxlen;
      
  if( c_info->s_alloc < s_len )
         {
         if( c_info->s_alloc > 0 ) 
                { free(c_info->s_buffer); }
         c_info->s_buffer = (void*) malloc(s_len);
         if( c_info->s_buffer == NULL )  Err_Hand(0, APPL_ERR_MALLOC);
         c_info->s_alloc = s_len;
         }
  if( c_info->r_alloc < r_len )
         {
         if( c_info->r_alloc > 0 ) 
                { free(c_info->r_buffer); }
         c_info->r_buffer = (void*) malloc(r_len);
         if( c_info->r_buffer == NULL )  Err_Hand(0, APPL_ERR_MALLOC);
         c_info->r_alloc = r_len;
         }

  MPI_Type_size(c_info->red_data_type,&s_size);
  
  if (s_size!=0) len = size/s_size;
  
  for (i= 0; i<len; i++)
    {
      ((float*)c_info->s_buffer)[i] =
          BUF_VALUE(c_info->rank,i);
      ((float*)c_info->r_buffer)[i] =  0.0;
    }
  for (i=len*sizeof(float) ; i<size; i++)
    {
      ((char*)c_info->s_buffer)[i] =  0 ;
      ((char*)c_info->r_buffer)[i] =  0 ;
    }
}

/********************************************************************/
void Free_ALL(struct comm_info* c_info, struct Bench** P_BList)
/********************************************************************

---------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
---------------------------------------------------------------------
In/Out     : c_info   | struct comm_info* | see comm_info.h 
                      |                   | contained pointers freed
             P_BList  | struct Bench**    | cotnained pointers freed
                      |                   | 
----------------------------------------------------------------------*/
{
  free(c_info->s_buffer);
  c_info->s_alloc=0;
  free(c_info->r_buffer);
  c_info->r_alloc=0;
  free(c_info->g_ranks);
  free(c_info->g_sizes);

  if( c_info->communicator != MPI_COMM_NULL &&
      c_info->communicator != MPI_COMM_WORLD )
  MPI_Comm_free(&c_info->communicator);

  Destruct_BList(P_BList);
  
}
